﻿using System.Reflection;
using HarmonyLib;
using System;
using System.Collections.Generic;
using UnityEngine;
using System.Reflection.Emit;
using System.Linq;
using System.Runtime.Serialization;
using Steamworks;

namespace Whitebrim.BetterSteeringWheel
{
    public class BetterSteeringWheel : Mod
    {
        Harmony harmonyInstance;

        public void Start()
        {
            harmonyInstance = new Harmony("com.whitebrim.bettersteeringwheel");
            harmonyInstance.PatchAll(Assembly.GetExecutingAssembly());
            Debug.Log("BetterSteeringWheel has been loaded!");
        }

        public void OnModUnload()
        {
            harmonyInstance.UnpatchAll();
            Debug.Log("BetterSteeringWheel has been unloaded!");
            Destroy(gameObject);
        }

        public void FixedUpdate()
		{
            NetworkMessage netMessage = RAPI.ListenForNetworkMessagesOnChannel(SteeringWheelPatch.CHANNEL_ID);
            if (netMessage != null)
            {
                Message message = netMessage.message;
                if (message.Type == (Messages)SteeringWheelPatch.MESSAGE_TYPE)
                {
                    Message_SteeringWheel_AdditionalData msg = message as Message_SteeringWheel_AdditionalData;
                    var steeringWheel = NetworkIDManager.GetNetworkIDFromObjectIndex<SteeringWheel>(msg.NetworkID_ObjectIndex);
                    steeringWheel.GetAdditionalData().SetData(msg.data);
                }
            }
        }
    }

    class SteeringWheelPatch
    {
        public static int CHANNEL_ID = 15006;
        public static Messages MESSAGE_TYPE = (Messages)1244;

        [HarmonyPatch(typeof(SteeringWheel), "OnIsRayed")]
        class OnIsRayed
        {
            private static void Postfix(SteeringWheel __instance, ref DisplayTextManager ___displayText, ref float ___targetLocalEulerZ, Semih_Network ___network, ref Network_Player ___localPlayer)
            {
                if (MyInput.GetButtonUp("Rotate"))
                {
                    var mouseLookXScript = ___localPlayer.PlayerScript.mouseLookXScript;
                    if (mouseLookXScript.isChild)
                        Traverse.Create(mouseLookXScript).Field("targetRotX").SetValue(Mathf.Clamp(mouseLookXScript.transform.localEulerAngles.y, mouseLookXScript.minimumX, mouseLookXScript.maximumX));
                    else
                        Traverse.Create(mouseLookXScript).Field("targetRotX").SetValue(Mathf.Clamp(mouseLookXScript.transform.eulerAngles.y, mouseLookXScript.minimumX, mouseLookXScript.maximumX));
                }

                if (!__instance.GetAdditionalData().isLocked)
                    ___displayText.ShowText("Press to LOCK rotation", MyInput.Keybinds["RMB"].MainKey, 2, 0, false);
                else
                    ___displayText.ShowText("Press to UNLOCK rotation", MyInput.Keybinds["RMB"].MainKey, 2, 0, false);
                ___displayText.ShowText("Press to Reset rotation", MyInput.Keybinds["LMB"].MainKey, 1, 0, false);

                if (MyInput.GetButtonDown("RMB"))
                {
					__instance.GetAdditionalData().isLocked = !__instance.GetAdditionalData().isLocked;
					if (!__instance.GetAdditionalData().isLocked)
                    {
                        float rotation = -___targetLocalEulerZ;
                        Message_SteeringWheel_Rotate message = new Message_SteeringWheel_Rotate(Messages.SteeringWheelRotate, ___network.NetworkIDManager, __instance.ObjectIndex, rotation);
                        if (Semih_Network.IsHost)
                        {
                            Traverse.Create(__instance).Method("Rotate", rotation).GetValue();
                            return;
                        }
                        ___network.SendP2P(___network.HostID, message, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
                    }
                }
                if (MyInput.GetButtonDown("LMB"))
                {
                    __instance.GetAdditionalData().isLocked = false;
                    float rotation = -___targetLocalEulerZ;
                    Message_SteeringWheel_Rotate message = new Message_SteeringWheel_Rotate(Messages.SteeringWheelRotate, ___network.NetworkIDManager, __instance.ObjectIndex, rotation);
                    if (Semih_Network.IsHost)
                    {
                        Traverse.Create(__instance).Method("Rotate", rotation).GetValue();
                        return;
                    }
                    ___network.SendP2P(___network.HostID, message, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
                }
				if (MyInput.GetButtonDown("RMB") || MyInput.GetButtonDown("LMB"))
				{
                    Message_SteeringWheel_AdditionalData message = new Message_SteeringWheel_AdditionalData(MESSAGE_TYPE, ___network.NetworkIDManager, __instance.ObjectIndex, __instance.GetAdditionalData());
                    RAPI.SendNetworkMessage(message, channel: CHANNEL_ID);
                }
			}
        }

        [HarmonyPatch(typeof(SteeringWheel), "OnRayExit")]
        class OnRayExit
        {
            private static void Postfix(SteeringWheel __instance, ref float ___targetLocalEulerZ, Semih_Network ___network)
            {
                if (!__instance.GetAdditionalData().isLocked)
                {
                    float rotation = -___targetLocalEulerZ;
                    Message_SteeringWheel_Rotate message = new Message_SteeringWheel_Rotate(Messages.SteeringWheelRotate, ___network.NetworkIDManager, __instance.ObjectIndex, rotation);
                    if (Semih_Network.IsHost)
                    {
                        Traverse.Create(__instance).Method("Rotate", rotation).GetValue();
                        return;
                    }
                }
            }
        }

        [HarmonyPatch(typeof(SteeringWheel), "Update")]
        class Update
        {
            private static IEnumerable<CodeInstruction> Transpiler(MethodBase original, IEnumerable<CodeInstruction> instructions)
            {
                var codes = new List<CodeInstruction>(instructions);
                for (int i = 0; i < codes.Count; i++)
                {
                    if (codes[i].opcode == OpCodes.Ldc_R4)
                    {
                        var strOperand = (float)codes[i].operand;
                        if (strOperand == 1)
                        {
                            codes[i].operand = 0.001f;
                            break;
                        }
                    }
                }
                return codes.AsEnumerable();
            }
        }

        [HarmonyPatch(typeof(SteeringWheel), "RestoreWheel")]
        class RestoreWheel
        {
            private static void Prefix(SteeringWheel __instance, RGD_SteeringWheel rgdWheel)
            {
                SteeringWheelAdditionalData value;
				if (SteeringWheelExtension.RGD_data.TryGetValue(rgdWheel, out value))
					__instance.AddData(value);
			}
        }
    }

    class RGD_SteeringWheelPatch
    {
        [HarmonyPatch(typeof(RGD_SteeringWheel), MethodType.Constructor, new Type[] { typeof(RGDType), typeof(SteeringWheel) })]
        class RGD_SteeringWheelConstructor1
        {
            private static void Prefix(RGD_SteeringWheel __instance, ref SteeringWheel steeringWheel)
            {
                __instance.AddData(steeringWheel.GetAdditionalData());
            }
        }

        [HarmonyPatch(typeof(RGD_SteeringWheel), MethodType.Constructor, new Type[] { typeof(SerializationInfo), typeof(StreamingContext) })]
        class RGD_SteeringWheelConstructor2
        {
            private static void Prefix(RGD_SteeringWheel __instance, ref SerializationInfo info)
            {
                try
                {
                    __instance.AddData(JsonUtility.FromJson<SteeringWheelAdditionalData>(info.GetString("AdditionalData")));
                }
				catch (Exception) { }
			}
        }

        [HarmonyPatch(typeof(RGD_SteeringWheel), "GetObjectData")]
        class GetObjectData
        {
            private static void Postfix(RGD_SteeringWheel __instance, ref SerializationInfo info)
            {
                SteeringWheelAdditionalData value;
				if (SteeringWheelExtension.RGD_data.TryGetValue(__instance, out value))
					info.AddValue("AdditionalData", JsonUtility.ToJson(value));
			}
        }
    }
}